#include "RobotGame.h"

#include <stdio.h>
#include <sys/types.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <queue>
#include <strstream>

// THINGS TO PONDER:
// 1. How to determine when other robots drop packages in places that would be worth investigating
// 2. When would it be a good idea to push someone else (if ever)?
// 3. What is the correct carry weight threshold for switching from base finding to package delivery?
int close(int fd);

#ifdef DEBUG
#define RG_DEBUG(__sz) printf("%s\n", __sz)
#define RG_RETURN(__num) do { printf("%s:%d - RC=%d\n", \
__FILE__, __LINE__, __num); return __num; } while (0)
#else
#define RG_DEBUG(__sz)
#define RG_RETURN(__num) return __num
#endif

RobotGame::RobotGame()
: _sockfd(-1), _width(-1), _height(-1), _id(-1), _maxCarry(-1), _curCarry(0), _money(-1), _score(0), _curState(stateUnknown)
{
}

RobotGame::~RobotGame()
{
	if (_sockfd >= 0)
	{
		// close(_sockfd);
		_sockfd = -1;
	}
}

int RobotGame::init(char *host, u_short port)
{
	sockaddr_in	serv_addr;
	bzero((char *)&serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;

	unsigned long inaddr = inet_addr(host);
	if (inaddr != INADDR_NONE)
		bcopy((char *)&inaddr, (char *)&serv_addr.sin_addr, sizeof(inaddr));
	else
		{
		hostent *hp = gethostbyname(host);

		if (hp == NULL)
			RG_RETURN(7);
		bcopy(hp->h_addr, (char *)&serv_addr.sin_addr, hp->h_length);
		}
	serv_addr.sin_port = htons(port);

	if ((_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		RG_RETURN(1);
	RG_DEBUG("Past socket - about to connect");
	if (connect(_sockfd, (sockaddr *)&serv_addr, sizeof(serv_addr)) < 0)
		RG_RETURN(2);

	RG_DEBUG("Past connect - about to send");
	// Now, send the "Player" string
	int nwritten = send(_sockfd, "Player\n", 7, 0);
//	int nwritten = send(_sockfd, "Login 2 2\n", 10, 0);
	if (nwritten < 0)
		RG_RETURN(3);

	RG_DEBUG("Past send - about to read board");
	if (readBoard())
		RG_RETURN(4);
	if (readConfig())
		RG_RETURN(5);
	if (readResponse())
		RG_RETURN(6);
	return 0;
}

int RobotGame::readBoard(void)
{
	string str;

	if (readline(str))
		RG_RETURN(1);	// Board dimensions
	string::size_type posSpace;

	if ((posSpace = str.find(" ")) == string::npos)
		RG_RETURN(2);
	
	_width = atoi(str.substr(0, posSpace).c_str());
	_height = atoi(str.substr(posSpace + 1,
				  str.length() - posSpace - 1).c_str());
	if (_width < 1 || _height < 1)
		RG_RETURN(3);
	_rgrgchBoard = new char *[_height];
	_rgrgiFlood = new int *[_height];
	for (int h = 0; h < _height; h++)
	{
		string strRow;
		_rgrgchBoard[h] = new char[_width];
		_rgrgiFlood[h] = new int[_width];
		if (readline(strRow))
			RG_RETURN(4);
		memcpy(_rgrgchBoard[h], strRow.c_str(), _width);
		for (int w = 0; w < _width; w++)
			if (_rgrgchBoard[h][w] == '@')
				_vecBases.push_back(new Location(w, h));
#ifdef DEBUG
		printf(strRow.c_str()); // UNDONE
#endif
	}
	return 0;
}

int RobotGame::readPackages(MMapWeightPkg &mmapWeightPkgCur)
{
	string str;
	if (readline(str))
		RG_RETURN(1);
	if (str[0] == '\n')
		return 0;

	string::size_type pos = 0;
	while (pos < str.length())
	{
		string::size_type posSpace = str.find(" ", pos);
		int id = atoi(str.substr(pos, posSpace - pos).c_str());
		pos = posSpace + 1;

		posSpace = str.find(" ", pos);
		int dstX = atoi(str.substr(pos, posSpace - pos).c_str());
		pos = posSpace + 1;

		posSpace = str.find(" ", pos);
		int dstY = atoi(str.substr(pos, posSpace - pos).c_str());
		pos = posSpace + 1;

		posSpace = str.find(" ", pos);
		if (posSpace == string::npos)
			posSpace = str.length() - 1;
		int weight = atoi(str.substr(pos, posSpace - pos).c_str());
		pos = posSpace + 1;

		Package *pkg = ensurePackage(id, dstX, dstY, weight);
		mmapWeightPkgCur.insert(MapIntPkg::value_type(weight, pkg));
	}
	return 0;
}

#define N_BIT (0x00000001)
#define E_BIT (0x00000002)
#define W_BIT (0x00000004)
#define S_BIT (0x00000008)
#define DIR_MASK (0x0000000F)

#define N_BASE (0x00000010)
#define E_BASE (0x00000020)
#define W_BASE (0x00000040)
#define S_BASE (0x00000080)
#define BASE_MASK (0x000000F0)
int RobotGame::bitValidDirs(int iX, int iY)
{
	int bitValid = 0;
	if (iX > 0)
		bitValid |= W_BIT;
	if (iX < _width - 1)
		bitValid |= E_BIT;
	if (iY > 0)
		bitValid |= S_BIT;
	if (iY < _height - 1)
		bitValid |= N_BIT;

	if (bitValid & N_BIT)
		switch(_rgrgchBoard[iY + 1][iX])	// Board upside down - Y increases in North dir
			{
			case '~':
			case '#':
				bitValid &= ~N_BIT;
				break;
			case '@':
				bitValid |= N_BASE;
				break;
			default:
				break;
			}
	if (bitValid & S_BIT)
		switch(_rgrgchBoard[iY - 1][iX])	// Board upside down - Y increases in North dir
			{
			case '~':
			case '#':
				bitValid &= ~S_BIT;
				break;
			case '@':
				bitValid |= S_BASE;
				break;
			default:
				break;
			}
	if (bitValid & E_BIT)
		switch(_rgrgchBoard[iY][iX + 1])
			{
			case '~':
			case '#':
				bitValid &= ~E_BIT;
				break;
			case '@':
				bitValid |= E_BASE;
				break;
			default:
				break;
			}
	if (bitValid & W_BIT)
		switch(_rgrgchBoard[iY][iX - 1])
			{
			case '~':
			case '#':
				bitValid &= ~W_BIT;
				break;
			case '@':
				bitValid |= W_BASE;
				break;
			default:
				break;
			}
	return bitValid;
}

int RobotGame::bestDir(Robot *pr, string &strDir)
{
	int iX = pr->getX() - 1;
	int iY = pr->getY() - 1;

	if (_curState == stateUnknown)
		{
		// Init flood array with -1's
		for (int h = 0; h < _height; h++)
			{
			int *pi = _rgrgiFlood[h];
			for (int w = 0; w < _width; w++)
				*pi++ = -1;
			}
	 	queue<Location *> qLoc;
		// Add destinations for packages we're carrying to the list
	 	set<int> &setPkgs = pr->getSetPkgs();
	 	for (set<int>::iterator iterPkg = setPkgs.begin(); iterPkg != setPkgs.end(); iterPkg++)
	 		{
	 		Package *ppkg = ensurePackage(*iterPkg, -1, -1, -1);	// Last args don't matter - package info will be there
	 		int iXPkg = ppkg->getDstX() - 1;
	 		int iYPkg = ppkg->getDstY() - 1;
	 		
	 		Location *pLoc = new Location(iXPkg, iYPkg);
	 		_rgrgiFlood[iYPkg][iXPkg] = 0;
	 		qLoc.push(pLoc);
	 		}
	 	
	 	// Add home bases to the queue if we don't have packages
	 	if (qLoc.empty())
	 		{
		 	vector<Location *>::iterator iter;
		 	for (iter = _vecBases.begin(); iter != _vecBases.end(); iter++)
		 		{
		 		if (iX != (*iter)->getX() || iY != (*iter)->getY())
		 			{
		 			Location *pLoc = new Location((*iter)->getX(), (*iter)->getY());
		 			_rgrgiFlood[pLoc->getY()][pLoc->getX()] = 0;
		 			qLoc.push(pLoc);
		 			}
		 		}
	 		}
	 	
	 	bool fFoundStart = false;
	 	while (!qLoc.empty() && !fFoundStart)
	 		{
	 		Location *pLoc = qLoc.front();
			int iCurFlood = _rgrgiFlood[pLoc->getY()][pLoc->getX()];
			if (iCurFlood >= 0)
				{
				int bitValid = bitValidDirs(pLoc->getX(), pLoc->getY());
				int *piNeighbor;
				int iXCur, iYCur;
				if (bitValid & N_BIT)
					{
					iXCur = pLoc->getX(); iYCur = pLoc->getY() + 1;
					if (iX == iXCur && iY == iYCur)
						fFoundStart = true;
					piNeighbor = &_rgrgiFlood[iYCur][iXCur];
					if (*piNeighbor < 0)
						{
						*piNeighbor = iCurFlood + 1;
						qLoc.push(new Location(iXCur, iYCur));
						}
					}
				if (bitValid & S_BIT)
					{
					iXCur = pLoc->getX(); iYCur = pLoc->getY() - 1;
					if (iX == iXCur && iY == iYCur)
						fFoundStart = true;
					piNeighbor = &_rgrgiFlood[iYCur][iXCur];
					if (*piNeighbor < 0)
						{
						*piNeighbor = iCurFlood + 1;
						qLoc.push(new Location(iXCur, iYCur));
						}
					}
				if (bitValid & E_BIT)
					{
					iXCur = pLoc->getX() + 1; iYCur = pLoc->getY();
					if (iX == iXCur && iY == iYCur)
						fFoundStart = true;
					piNeighbor = &_rgrgiFlood[iYCur][iXCur];
					if (*piNeighbor < 0)
						{
						*piNeighbor = iCurFlood + 1;
						qLoc.push(new Location(iXCur, iYCur));
						}
					}
				if (bitValid & W_BIT)
					{
					iXCur = pLoc->getX() - 1; iYCur = pLoc->getY();
					if (iX == iXCur && iY == iYCur)
						fFoundStart = true;
					piNeighbor = &_rgrgiFlood[iYCur][iXCur];
					if (*piNeighbor < 0)
						{
						*piNeighbor = iCurFlood + 1;
						qLoc.push(new Location(iXCur, iYCur));
						}
					}
				}
			qLoc.pop();
	 		delete pLoc;
	 		}
		// Delete remaining locations from queue
	 	while (!qLoc.empty())
	 		{
	 		delete qLoc.front();
	 		qLoc.pop();
	 		}
	 	if (fFoundStart)
	 		_curState = stateToBase;
		}
	
	if (_curState != stateUnknown)
		{
		int iMin = 0x7FFFFFFF;
		char chDir = '.';

		int bitValid = bitValidDirs(iX, iY);
#ifdef DEBUG
		printf("bitValid == %x\n", bitValid);
#endif
		if (bitValid & N_BIT)
			{
			int iCur = _rgrgiFlood[iY + 1][iX];
			if (iCur >= 0 && iCur < iMin)
				{
				iMin = iCur;
				chDir = 'N';
				}
			}
		if (bitValid & S_BIT)
			{
			int iCur = _rgrgiFlood[iY - 1][iX];
			if (iCur >= 0 && iCur < iMin)
				{
				iMin = iCur;
				chDir = 'S';
				}
			}
		if (bitValid & W_BIT)
			{
			int iCur = _rgrgiFlood[iY][iX - 1];
			if (iCur >= 0 && iCur < iMin)
				{
				iMin = iCur;
				chDir = 'W';
				}
			}
		if (bitValid & E_BIT)
			{
			int iCur = _rgrgiFlood[iY][iX + 1];
			if (iCur >= 0 && iCur < iMin)
				{
				iMin = iCur;
				chDir = 'E';
				}
			}
		if (chDir == '.')
			RG_RETURN(7);
		strDir += chDir;
#ifdef DEBUG
		printf("iMin=%d, heading %s\n", iMin, strDir.c_str());
#endif
		return 0;
		}

	RG_RETURN(5);
}

int RobotGame::writeMove(MMapWeightPkg &mmapWeightPkgCur)
{
	int bid = 1;
	Robot *pr = ensureRobot(_id);
	string strMove;
	
	int iX = pr->getX() -1; int iY = pr->getY() - 1;

	// If we're at a home base, record that we've been here
 	vector<Location *>::iterator iter;
 	for (iter = _vecBases.begin(); iter != _vecBases.end(); iter++)
 		{
 		if (iX == (*iter)->getX() && iY == (*iter)->getY())
 			{
 			if (mmapWeightPkgCur.size() == 0) // Only if no more packages are here
 				{
#ifdef DEBUG
 				printf("Discarding base at (%d, %d)\n", iX+1, iY+1);
#endif
 				_vecBasesVisited.push_back(*iter);
 				_vecBases.erase(iter);
 				}
 			_curState = stateUnknown;
 			break;
 			}
 		}
 	
 	bool fMove = true;
 	// If are carrying packages that should be dropped, do it
	// Also add destinations for packages we're carrying to the list
 	set<int> &setPkgs = pr->getSetPkgs();
 	for (set<int>::iterator iterPkg2 = setPkgs.begin(); iterPkg2 != setPkgs.end(); iterPkg2++)
 		{
 		Package *ppkg = ensurePackage(*iterPkg2, -1, -1, -1);	// Last args don't matter - package info will be there
 		int iXPkg = ppkg->getDstX() - 1;
 		int iYPkg = ppkg->getDstY() - 1;
 		if (iX == iXPkg && iY == iYPkg)
 			{
 			if (fMove)
 				{
 				fMove = false;
 				strMove += "1 Drop";
 				_curState = stateUnknown;	// This will force us to pick a new destination next time
 				}
 			strMove += " ";
 			strstream ss;
 			ss << *iterPkg2 << '\0';
 			strMove += ss.str();
 			}
 		}
 	
 	// If we have packages here we can pick up, do it!
 	if (fMove)
 		{
	 	MMapWeightPkg::iterator iterPkg;
	 	int maybeCarry = _curCarry;
	 	unsigned int iPkg;
	 	for (iPkg = 0, iterPkg = mmapWeightPkgCur.end(); iPkg < mmapWeightPkgCur.size(); iPkg++)
	 		{
	 		iterPkg--;
	 		if (_maxCarry - maybeCarry >= iterPkg->first)
	 			{
	 			if (fMove)
	 				{
	 				fMove = false;
	 				// UNDONE: If other robots are at this location, up the bid
	 				strMove += "1 Pick";
	 				}
	 			strMove += " ";
	 			strstream ss;
	 			ss << iterPkg->second->getId() << '\0';
	 			strMove += ss.str();
	 			maybeCarry += iterPkg->first;
	 			}
	 		}
 		}
 	
 	// No pickups or drops, we can move then
	if (fMove)
		{
		strMove += "1 Move ";
		string strDir;
		if (bestDir(pr, strDir))
			RG_RETURN(1);
		strMove += strDir;
		}
	strMove += '\n';
#ifdef DEBUG
	printf(strMove.c_str()); // UNDONE
#endif
	// Now, send the move string
	int nwritten = send(_sockfd, strMove.c_str(), strMove.length(), 0);
	if (nwritten < 0)
		RG_RETURN(2);
	_money--;
	return 0;
}

Package *RobotGame::ensurePackage(int id, int dstX, int dstY, int weight)
{
	MapIntPkg::iterator iter;
	iter = _mapIdPkg.find(id);
	if (iter != _mapIdPkg.end())
		return iter->second;
	Package *pkg = new Package(id, dstX, dstY, weight);
	_mapIdPkg.insert(MapIntPkg::value_type(id, pkg));
	return pkg;
}

int RobotGame::readConfig(void)
{

	string str;

	if (readline(str))
		RG_RETURN(1);	// <id> <maxCarry> <money>

	string::size_type posSpace, posSpace2;
	if ((posSpace = str.find(" ")) == string::npos)
		RG_RETURN(2);
	
	_id = atoi(str.substr(0, posSpace).c_str());
	if ((posSpace2 = str.find(" ", posSpace + 1)) == string::npos)
		RG_RETURN(3);
	
	_maxCarry = atoi(str.substr(posSpace + 1,
				    posSpace2 - posSpace - 1).c_str());
	_money = atoi(str.substr(posSpace2 + 1,
				 str.length() - posSpace2 - 1).c_str());
#ifdef DEBUG
	printf("Config: _id=%d _maxCarry=%d _money=%d\n", _id, _maxCarry, _money);
#endif
	return 0;
}

Robot *RobotGame::ensureRobot(int id)
{
	MapIdRobot::iterator iter = _mapIdRobot.find(id);
	if (iter != _mapIdRobot.end())
		return iter->second;
	// We need to create a new robot and insert
	Robot *pr = new Robot(id, this);
	_mapIdRobot.insert(MapIdRobot::value_type(id, pr));
	return pr;
}

int Robot::handleResponse(string &str)
{
	string::size_type pos = 0;
       	string::size_type increment = 2;

	while (pos < str.length())
	{
		switch(str[pos])
		{
		case 'X':
		{
			RG_DEBUG("Robot: X/Y");
			pos += 2;
			string::size_type posSpace = str.find(" ", pos);
			_x = atoi(str.substr(pos, posSpace - pos).c_str());
			pos = posSpace + 3; // Skip Y and space
			posSpace = str.find(" ", pos);
			if (posSpace == string::npos)
				posSpace = str.length() - 1;
			_y = atoi(str.substr(pos, posSpace - pos).c_str());
			pos = posSpace - 1; // Adjust for increment
		}
		break;
		case 'N':
			RG_DEBUG("Robot: North");
			_y++;
			break;
		case 'S':
			RG_DEBUG("Robot: South");
			_y--;
			break;
		case 'E':
			RG_DEBUG("Robot: East");
			_x++;
			break;
		case 'W':
			RG_DEBUG("Robot: West");
			_x--;
			break;
		case 'P':
		case 'D':
		{
			RG_DEBUG("Robot: Pick or Drop");
			bool fAdd = str[pos] == 'P';
			pos += 2;
			string::size_type posSpace = str.find(" ", pos);
			if (posSpace == string::npos)
				posSpace = str.length() - 1;
			int pkg =
				atoi(str.substr(pos, posSpace - pos).c_str());
			if (fAdd)
				_setPkgs.insert(pkg);
			else
				_setPkgs.erase(pkg);
			// If this robot is our robot, update our current carry weight
			if (_id == _prg->getId())
				if (_prg->pickOrDropPackage(pkg, fAdd))
					RG_RETURN(2);
			pos = posSpace - 1; // Adjust for increment
		}
		break;
		case '\n':
			// DO NOTHING
			break;
		default:
			RG_DEBUG(str.c_str());
			// Only return error if the offending response was for us
//			if (_id == _prg->getId())
//				RG_RETURN(1);
//			else
			return 0;
		}
		pos += increment;
	}
	return 0;
}

int RobotGame::readResponse(void)
{
	string str;

	if (readline(str))
		RG_RETURN(1);
	// RG_DEBUG(str.c_str());

	string::size_type posPound;
	posPound = str.find("#");
	while (posPound != string::npos)
	{
		string::size_type posSpace, posPound2, posEnd;
		posSpace = str.find(" ", posPound);
		int robotId =
			atoi(str.substr(posPound + 1,
					posSpace - posPound - 1).c_str());
		Robot *pr = ensureRobot(robotId);

		posPound2 = str.find("#", posPound + 1);
		if (posPound2 != string::npos)
			posEnd = posPound2;
		else
			posEnd = str.length();

		string strOneRobot = str.substr(posSpace + 1,
						posEnd - posSpace - 1);
		if (pr->handleResponse(strOneRobot))
			RG_RETURN(2);
#ifdef DEBUG
		printf("%d: X=%d Y=%d\n", robotId, pr->getX(), pr->getY());
#endif
		posPound = posPound2;
	}
		
	return 0;
}

int RobotGame::pickOrDropPackage(int id, bool fPick)
{
	MapIntPkg::iterator iter = _mapIdPkg.find(id);
	if (iter == _mapIdPkg.end())
		RG_RETURN(1);
	int weight = iter->second->getWeight();
	if (fPick)
		_curCarry += weight;
	else
		{
		_curCarry -= weight;
		
		// See if dropping this package increased our score
		Robot *pr = ensureRobot(_id);
		if (pr->getX() == iter->second->getDstX() && pr->getY() == iter->second->getDstY())
			{
			_score += weight;
#ifdef DEBUG
			printf("Scored!  Score now is %d\n", _score);
#endif
			}
		}
#ifdef DEBUG
	printf("_curCarry changed to %d\n", _curCarry);
#endif
	return 0;
}

int RobotGame::play(void)
{
	bool fDone = false;
	do
		{
		MMapWeightPkg mmapWeightPkgCur;
		
		if (readPackages(mmapWeightPkgCur))
			fDone = true;
		else if (writeMove(mmapWeightPkgCur))
			fDone = true;
		else if (readResponse())
			fDone = true;
		} while (!fDone);
	return 0;
}

int RobotGame::readline(string &str)
{
	int n = 0, rc;
	char c;

	for (;;n++)
	{
		if ((rc = recv(_sockfd, &c, 1, 0)) == 1)
		{
			str += c;
			if (c == '\n')
			{
				//	RG_DEBUG(str.c_str());
				return 0;
			}
		}
		else if (rc == 0)
			if (n == 0)
				RG_RETURN(1);
			else
				return 0;
		else
			RG_RETURN(2);	// ERROR
	}
}

