#ifndef _ROBOTGAME_H
#define _ROBOTGAME_H

#include <string>
#include <map>
#include <set>
#include <vector>

class Robot
{
 public:
	Robot(int id, class RobotGame *prg) : _prg(prg), _id(id), _x(-1), _y(-1) {}
	int handleResponse(string &str);
	int getX() { return _x; }
	int getY() { return _y; }
	int getId() { return _id; }
	set<int> &getSetPkgs() { return _setPkgs; }
	
 private:
	class RobotGame *_prg;
	int _id;
	int _x, _y;
	set<int> _setPkgs;
};
typedef map<int, Robot *> MapIdRobot;

class Package
{
public:
	Package(int id, int dstX, int dstY, int weight)
		: _id(id), _dstX(dstX), _dstY(dstY), _weight(weight) {}
	
	int getId(void) { return _id; }
	int getDstX(void) { return _dstX; }
	int getDstY(void) { return _dstY; }
	int getWeight(void) { return _weight; }
private:
	int _id, _dstX, _dstY, _weight;
};
typedef map<int, Package *> MapIntPkg;
typedef multimap<int, Package *> MMapWeightPkg;

class Location
{
public:
	Location(int x, int y) : _x(x), _y(y) {}
	int getX() { return _x; }
	int getY() { return _y; }
private:
	int _x, _y;
};
class RobotGame
{
 public:
 	typedef enum {
 		stateUnknown = 0,
 		stateToBase,
 		stateToDestination
 		} RobotState;
 	
	RobotGame();
	~RobotGame();

	int init(char *host, u_short port);
	int play(void);

	int getId(void) { return _id; }
	int pickOrDropPackage(int id, bool fPick);
	int getScore() { return _score; }
	int getMoney() { return _money; }
	
 private:
	int readline(string &str);

	int readBoard(void);
	int readConfig(void);
	int readResponse(void);
	int readPackages(MMapWeightPkg &mapWeightPkgCur);
	int writeMove(MMapWeightPkg &mapWeightPkgCur);
	int bestDir(Robot *pr, string &strDir);
	int bitValidDirs(int iX, int iY);
	
	Robot *ensureRobot(int id);
	Package *ensurePackage(int id, int dstX, int dstY, int weight);
	
	int	_sockfd;
	int 	_width, _height;
	int 	_id;
	int	_maxCarry;
	int	_curCarry;
	int	_money;
	int	_score;
	RobotState	_curState;
	vector<Location *> _vecBases;
	vector<Location *> _vecBasesVisited;
	MapIdRobot _mapIdRobot;
	char ** _rgrgchBoard;
	int **_rgrgiFlood;
	MapIntPkg _mapIdPkg;	// Key is pkg id
};

#endif // _ROBOTGAME_H

